// ----------------------------------------------------------------------------
// Copyright (C) 2018
//              David Freese, W1HKJ
//
// This file is part of pskmeter.
//
// pskmeter is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 3 of the License, or
// (at your option) any later version.
//
// pskmeter is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
// ----------------------------------------------------------------------------
#include <iostream>
#include <FL/Fl.H>

#include "support.h"

extern void display_data();

Cpskdata::Cpskdata() {
	mode = PSK31;
	ncols = 4;
	nrows = NUMPTS / ncols;
	sizeText = nrows * (5 * ncols + 1);
	szTEXT = new char[sizeText + 1];
	szRMS = new char[40];
	szIMD = new char[40];
	szPEAK = new char[40];
	for (int i = 0; i <= NUMPTS; i++) fVals[i] = 0.0;
	for (int i = 0; i < NUMPTS; i++) sample[i] = 0.0;
	imd = rms = peak = 0.0;
	no_signal = true;
	for (int i = 0; i < NUMPTS; i++) {
		sinTable[i] = sin( TWOPI * i / NUMPTS );
		cosTable[i] = cos( TWOPI * i / NUMPTS );
	}
// default configuration values
	showrawvals = true;

	alpha = 0.5;    // simple low pass filter applied to raw data
	R1qro = 39.0;
	R1qrp = 6.8;
	qrp = false;
	ratio = 1 + R1qro / 2.2;

	AtoD = A2D;

//	calmode = false;

	powerconv = 1.0 / (2 * 50 * sqrt(2));

// These starting values are valid for QRO
// at idle
// Prms = ((Vp*ratio)^2)/(2*sqrt(2)*R) (rf power ..... not audio power)
// where Vp = measured peak audio voltage
//       ratio = (1 + R1/R2);
//       R = real part of antenna impedance, nominally 50 ohms

//	targetPower = 15.0;
//	targetVoltage = sqrtf(targetPower*powerconv)/ratio;

	overdrive1 = 0.002;
	overdrive2 = 0.01;
	minval = 5.0 * 8 / 1024;

	signal_start = false;
	no_signal = true;
	valid_signal = false;
	_ready = true;
}

void Cpskdata::sanitize () {
	int minloc = 0;
	double diff = 0;
	min = 5.0;
	max = 0;

// find the min and max values and sample location of min value  
	for (int i = 0; i < NUMPTS; i++) {
		if (sample[i] > max) 
			max = sample[i];
		if (sample[i] < min) {
			min = sample[i];
			minloc = i;
		}
	}

// if (max - min) is less than threshold then data array does not contain
// a null point ... sample occurred in middle of series of 1 bits
// if (max < threshold) then NOT TRANSMITTING so reset all computations
// if (min > minval) then a true null was not sampled
// sample probably occurred at the leading or trailing edge of a series of 1 bits

	valid_signal = false;

	if ((max-min) < minval || max <= minval) {
		for (int i = 0; i < NUMPTS; i++) sample[i] = 0;
		no_signal = true;
		rms = 0.0;
		peak = 0.0;
		imd = 0.0;
		for (int i = 0; i < 7; i++) q[i] = 0.0;
		qsum = 0.0;
		signalPower = 0.0;
		return;
	}

	if ((max - min) < (3 * max / 4))  // did not sample a valid cross-over point
		return;

	if (no_signal) {
		signal_start = true;
		no_signal = false;
	}

	int temp[NUMPTS];
	for (int i = minloc, j = 0; i < NUMPTS; i++, j++)
		temp[j] = sample[i];
	for (int i = 0, j = NUMPTS - minloc; i < minloc; i++, j++)
		temp[j] = sample[i];

	diff = 0;
	for (int i = 0; i < NUMPTS; i++) {
		sample[i] = temp[i];
		if (i && (fabs(sample[i] - sample[i-1]) > diff))
			diff = fabs(sample[i] - sample[i-1]);
	}

	if (minloc > 48)
		return;

	if (diff > max / 2)
		return;

	no_signal = false;
	valid_signal = true;

	return;
}

// low pass filter raw data and create pseudo signal tables
// perform dft and statistical analysis
void Cpskdata::filter () {
	if (no_signal) peak = 0;
	else if (signal_start) peak = max;
	else peak = alpha * peak + (1-alpha) * max; 

	if (mode == PSK31) {
		if (no_signal)
			for (int i = 0; i <= NUMPTS; i++) fVals[i] = dftVals[i] = 0;
		else if (signal_start) {
			for (int i = 0; i < NUMPTS; i++)
				fVals[i] = dftVals[i] = sample[i];
			fVals[NUMPTS] = dftVals[NUMPTS] = fVals[NUMPTS - 1];
			signal_start = false;
		} else {
			for (int i = 0; i < NUMPTS; i++)
				fVals[i] = sample[i];
			fVals[NUMPTS] = fVals[NUMPTS - 1];

			if (valid_signal) {
				for (int i = 0; i <= NUMPTS / 4; i++)
					dftVals[i] = alpha * dftVals[i] + (1.0 - alpha) * fVals[i];
				for (int i = 0; i < NUMPTS / 4; i++)
					dftVals[NUMPTS/2 - i] = dftVals[i];
				for (int i = 1; i <= NUMPTS / 2; i++)
					dftVals[NUMPTS/2 + i] = -dftVals[i];
			}
		}
	} else {
		if (no_signal)
			for (int i = 0; i <= NUMPTS; i++) fVals[i] = 0;
		else if (signal_start) {
			for (int i = 0; i < NUMPTS; i++)
				fVals[i] = sample[i];
			signal_start = false;
		} else {
			for (int i = 0; i <= NUMPTS / 8; i++)
				fVals[i] = alpha * fVals[i] + (1.0 - alpha) * sample[i];
			for (int i = 0; i <= NUMPTS / 8; i++) {
				fVals[NUMPTS/2 + i] = fVals[i];
				fVals[NUMPTS/4 + i] = fVals[NUMPTS / 2 + NUMPTS / 4 + i] = -fVals[i];
			}
			for (int i = 0; i < NUMPTS / 8; i++) {
				fVals[NUMPTS / 4 - 1] = fVals[NUMPTS / 2 + NUMPTS / 4 - i] = fVals[i];
				fVals[NUMPTS / 2 - i] = fVals[NUMPTS - i] = fVals[i];
			}
		}
	}

}

// dft - discrete fourier transform analysis
void Cpskdata::dft()
{
	int si; // sample increment
	double a = 0;
	double b = 0;
	double denom = NUMPTS / 2.0;
	denom *= denom;

	for (int i = 0; i < 7; i += 2)
		q[i] = 0.0;

	if (!valid_signal) return;

	signalPower = 0.0;
	for (int j = 0; j < 7; j += 2) { // fundamental, 3rd, 5th & 7th harmonics
		a = b = 0.0;
		for (int i = 0; i < NUMPTS; i++) {
			if (mode == PSK31)
				si = ((j+1)*i) % NUMPTS;
			else
				si = (2*(j+1)*i) % NUMPTS;
			a += dftVals[i]*sinTable[si];
			b += dftVals[i]*cosTable[si];
		}
		q[j] = (a*a + b*b) / denom;
	}

	qsum = 0;
	for (int i = 0; i < 7; i += 2) {
		qsum += q[i]; // sum of all odd harmonic powers
		if (i > 0) q[i] /= q[0];
	}
	q[0] = 1.0;

	signalPower = qsum * powerconv;

	if (signalPower < 0.1)
		for (int i = 0; i < 7; q[i++] = 0);

}

void Cpskdata::data(double *rawdata) {
	_ready = false;
	for (int i = 0; i < NUMPTS; i++)
		sample[i] = rawdata[i] * ratio;

	sanitize();
	filter();
	dft();

	Fl::awake(display_data);
}

const char *Cpskdata::szRawData() {
	char szt[250];
	static string szTEXT;
	szTEXT.clear();
	for (int n = 0; n < NUMPTS/2; n += 4) {
		sprintf(szt, "%8d %8d %8d %8d\n",
			(int)(A2DSIZE * sqrtf(fVals[ n ])/ratio), 
			(int)(A2DSIZE * sqrtf(fVals[n+1])/ratio), 
			(int)(A2DSIZE * sqrtf(fVals[n+2])/ratio), 
			(int)(A2DSIZE * sqrtf(fVals[n+3])/ratio));
		szTEXT.append(szt);
	}
	return szTEXT.c_str();
}

const char *Cpskdata::szFilteredData() {
	char szt[250];
	static string szTEXT;
	szTEXT.clear();
	for (int n = 0; n < NUMPTS / 2; n += 4) {
		sprintf( szt, "%6.1f %6.1f %6.1f %6.1f\n",
			sqrtf(fVals[ n ]),
			sqrtf(fVals[n+1]),
			sqrtf(fVals[n+2]),
			sqrtf(fVals[n+3]));
		szTEXT.append(szt);
	}
	return szTEXT.c_str();
}

const char * Cpskdata::szPeak() {
	if (no_signal) 
		sprintf(szRMS, "------");
	else
		sprintf(szPEAK, "%4.1f", peak);
	return szPEAK;
}

const char *Cpskdata::szImd() {
	if (no_signal) 
		sprintf(szTEXT, "------");
	else
		sprintf(szTEXT, "%4.1f", 10.0*log10(q[2]/q[0])); // ratio of 3rd to 1st
	return szTEXT;
}

int Cpskdata::IMDstat() {
	if (no_signal)
		return 0;
	if (q[2] > overdrive1)
		return 1;
	if (q[2] > overdrive2)
		return 2;
	return 0;
}

double Cpskdata::SignalPwr() {
	if (no_signal) return 0.0;
	return signalPower;
}

const char *Cpskdata::szPower() {
	if (no_signal) 
		sprintf(szTEXT, "------");
	else
		sprintf(szTEXT,"%4.1f", signalPower);
	return szTEXT;
}

const char *Cpskdata::szSpectralLines() {
	sprintf(szTEXT,"%7.5f,%7.5f,%7.5f,%7.5f\n%7.5f,%7.5f,%7.5f",
	q[0], q[1], q[2], q[3], q[4], q[5], q[6]);
	return szTEXT;
}
